// PROJECT:       Micro-Manager
// SUBSYSTEM:     MMCore
//
// COPYRIGHT:     University of California, San Francisco, 2014,
//                All Rights reserved
//
// LICENSE:       This file is distributed under the "Lesser GPL" (LGPL) license.
//                License text is included with the source distribution.
//
//                This file is distributed in the hope that it will be useful,
//                but WITHOUT ANY WARRANTY; without even the implied warranty
//                of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//
//                IN NO EVENT SHALL THE COPYRIGHT OWNER OR
//                CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
//                INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES.
//
// AUTHOR:        Mark Tsuchida

#pragma once

#include "CameraInstance.h"
#include "ShutterInstance.h"
#include "StageInstance.h"
#include "XYStageInstance.h"
#include "StateInstance.h"
#include "SerialInstance.h"
#include "GenericInstance.h"
#include "AutoFocusInstance.h"
#include "ImageProcessorInstance.h"
#include "SignalIOInstance.h"
#include "MagnifierInstance.h"
#include "SLMInstance.h"
#include "GalvoInstance.h"
#include "HubInstance.h"
